"""

This file contains simple input validation and type conversion for
the instrument forms.

The input validation checks that you aren't trying to sneak something
by to subvert the system.  For example, after validation, an instrument
name can only ever contain letters.  Even if you use the instrument
name in os.system(), you don't have to worry about the user submitting
an instrument named "; rm -rf / ; " and fooling your program into removing
files from your system.

The entry point is instrument_input_conversion().

All request forms share a single name space.  If an instrument has form
field "xyz", it has the same meaning as "xyz" for every other instrument.
The dict named "converter" contains one function for each input field.
That function converts the input value and raises an exception if the
value doesn't look right or can't be converted to the right type.

In principle, the user never sees an exception because javascript in
the input form will prevent a bad form value being submitted.
(This is not fully implemented. 2009-08-19)

There are several functions defined for use in converter[].  You can use
one of them directly, or you can use a lambda function.  If the expression
gets complicated, you would want to pull it out into a separate function
anyway, but we make extensive use of lambda function to verify lists:

converter = {

    'sample1':  alnum,
            # Tests that the field is an alphanumeric string

    'sample2':  float_range,
            # the field is converted to floating point

    'sample3':  lambda v, n : float_range( v, n, min=1.0, max=10.0 ),
            # the field is floating point with min/max values

    'sample4':  lambda v, n : int_range( v, n, min=0, max=10 ),
            # same thing as integer

    'sample5':  lambda v, n : list_int( v, n, [ 1, 2, 4, 8, 16 ] ),
            # field is integer, but only one of the listed values

    'sample6':  lambda v, n : list_str( v, n, [ "a", "b", "hello" ]),
            # field is string, but only one of the listed values

    'sample7':  declination,
    'sample8':  right_ascension,
            # these return strings because that is what other parts of
            # the system expect.  Should they be something else?

There is a thing at the end of this file that checks that all the needed
fields are listed in converter[].  It runs on import if debugging is
turned on.

"""
import os, re, sys

import pysynphot
from pandeia.engine.helpers.bit.pyetc_util import dynamic_import, log
CFG_DEBUG = log.VERBOSE_FLAG # no need for full web config here, just use this


valid_instrument_name = re.compile('^[a-z0-9]+$')
valid_science_mode = re.compile('^[A-Za-z_]+$')

instrument_modules = { }
################################################################################

class InputConversionError(ValueError):
    """An exception for bad inputs which should be displayed as other than server
    errors.
    """
    pass

class InputConversionHackDetect(Exception):
    '''An exception to raise if the input looks like an attack

    In this case, we just let it bubble up to the user as "internal error"
    without giving any clues.  Developers can see it in the log file.
    '''
    pass

def handle_file_uploads(request, request_dir, upload_list):
    """Check uploaded files and copy to request directory.
    """
    # if there are any uploaded files, store them in request_dir.
    #
    # For an upload filed named 'Abc', the request will contain a key
    # 'Abc' that contains the name the user gave the file, and 'AbcLocal'
    # that contains the base name of the actual file on disk.
    #
    # The file name on disk is the form field name, plus any extension
    # that the user provided IF we recognize the extension.
    # For example, if the form contains "fUploadFile", and the user
    # uploads "xyz.fits", request_dir will contain "fUploadFile.fits".
    #

    # upload_list is a list of the uploaded files that the input verifier
    # saw in the form and considered valid upload names.  Therefore, we
    # know that they are all there.  The input verifier did not check
    # the files, because we can't fully accept the upload until after
    # creating the etc_id.  So, we verify it in handle_upload()
    #
    updates = {}
    for x in upload_list :
        updates[x] = request.FILES[x].name
        # xxxLocal is the file name where we stored the upload for field xxx
        # it is relative to request_dir
        # The actual file name may end with ".fits" or ".txt" or "".
        updates[x+'Local'] = handle_upload(request, request_dir, x)
    return updates


##

# see use of upload_helper in handle_upload just below
def upload_helper( request, fieldname, request_dir ) :

    # the name of the form field will be the name of the file on our disk...
    f = request.FILES[fieldname]
    filename = str(fieldname)

    # ...maybe with .fits or .txt added
    for x in [ '.fits', '.txt' ] :
        if f.name.endswith(x) :
            filename = filename + x
            break

    # collect the data into the file.
    # (ok to do now - we would not be here if it is not a field we expect.)
    fullpath = os.path.join(request_dir, filename )
    where = open(fullpath,"wb+")
    for chunk in f.chunks():
        where.write(chunk)
    where.close()

    #
    return filename, fullpath


def handle_upload(request, request_dir, fieldname ) :
    """Stores an uploaded file from the web form into the data directory.

    request - the django request object

    fieldname - the name of the field in the input form.

    request_dir - the name of the data directory to store the file in; the
        file name will be the same as the fieldname, but with recognized
        extentions copied from the uploaded name.

    recgonized extentions are:
        .fits
        .txt

    Security note:  This function guards against attacks where there
    user selects an arbitrary field name to get around our file system
    limitations.  It only creates files for those uploadable fields
    that we recognize.  Uploads of other files are not stored.

    """
    if fieldname in request.FILES :

        if fieldname == 'fUploadFile' :
            filename, fullpath = upload_helper( request, fieldname, request_dir )
            try:   # Load either a fits file or an ASCII file.
                   # In any case,  let pysynphot figure it out.
                   # If the file loads,  it's valid.  Discard result.
                pysynphot.FileSpectrum(fullpath)
            except Exception as e:
                # It was an invalid file, but we leave it there for the help desk to look at
                # os.remove(fullpath)
                msg = "Invalid format for uploaded file " + repr(str(request.FILES[fieldname].name))

                # provide more detail if we have any
                if isinstance(e,pysynphot.exceptions.TableFormatError):
                    msg = msg + '\n' + '\n'.join(e.args)
                    raise InputConversionError(msg)

                # raise something even if we don't recognize the problem
                raise

        ## add more uploadable file types here
        # if fieldname == 'whatever' :
        #   filename, fullpath = upload_helper( request, fieldname, request_dir )
        #   ...validate the file...

        else :
            # did not save the file
            raise InputConversionHackDetect('Do not know how to validate file upload %s'%fieldname)

        #
        return filename

    return None


###############################################################################
def convert_dictionary( indict, flist=None):
    """ Alternate entry point for converting a vanilla dictionary
    rather than an actual request.  Used during testing.

    indict:  dictionary
    flist:   optional list of files to be treated as upload files
    """
    strdict = {}
    for key, value in list(indict.items()):
        strdict[key] = str(value)
    request = FakeRequest(strdict, flist)
    fdict, upload_list = instrument_input_conversion( request,
                                                      indict['instrument'],
                                                      indict['science_mode'])
    return fdict, upload_list

class FakeRequest(object):
    """Helper class: used to lie to instrument_input_conversion during testing

    This object looks enough like a django request that we can process
    it through the input conversion, but we don't have to make a complete
    django request object.
"""

    def __init__(self, indict, flist):
        self.POST = indict
        if flist is None:
            self.FILES =[]
        else:
            self.FILES = flist


################################################################################
def instrument_input_conversion( request, instrument, science_mode) :

    # Input is
    #   request: the django request, which contains the form fields that
    #       were submitted.  These are the fields to be validated.
    #   instrument, science_mode: name of the instrument/science_mode
    #       that this form is for; we use this information to choose
    #       what fields to validate.
    #
    # Return is a tuple of ( converted_dict, upload_info ), where
    #   converted_dict: is a dictionary of validated and type-converted
    #       form values, indexed by form field name.
    #   upload_info: is data to hand to handle_uploads() when it is time
    #       to extract uploaded files from the request
    #
    # If any of the fields fail validation, we raise InputConversionError().
    #

    # This part duplicates tests that are already implicit in the django
    # urls.py file, but it doesn't cost much to do it again here and it
    # gives us some protection against programmer errors in future
    # changes to urls.py

    if not valid_instrument_name.match(instrument) :
        raise InputConversionHackDetect('Invalid instrument name %s'%instrument)
    if not valid_science_mode.match(science_mode) :
        raise InputConversionHackDetect('Invalid science mode %s'%science_mode)

    # The list of which fields to expect from each form is stored in
    # the instruments package for that instrument.  Construct the
    # name of the relevant module and import it.

    # Note that the existence of the package pyetc.instruments.XXX_YYY is
    # sufficient validation that XXX/YYY is a valid instrument and science mode.

    key = '%s_%s'%(instrument, science_mode)

    if not key in instrument_modules :
        try :
            m = dynamic_import("pandeia.engine.helpers.bit.instruments.hst.%s.web.form_fields_%s_%s" % (
                instrument, instrument, science_mode))
        except ImportError :
            raise InputConversionHackDetect('invalid instrument/science mode: %s %s'%(
                instrument, science_mode))
        instrument_modules[key] = m
    else :
        m = instrument_modules[key]

    # Now we have the list of fields that we expect to find in the form.
    field_list = m.field_list

    # This dictionary is a list of all the input fields that we actually
    # received.  As we recognize each field, we remove it.  At the end,
    # this dictionary is empty.
    original_inputs = { }
    original_inputs.update(request.POST)

    # This is the result dictionary that contains the translated
    # form inputs.
    fdict = { }

    # This is be the list of uploaded files.  We only list the uploaded
    # files that we actually expect for this form.  (If somebody uploads
    # something else, we never retrieve it.)
    upload_list = [ ]

    # a place to collect all the exceptions that this input form raises.
    # You could just raise an exception on the first bad field, but
    # that sometimes makes it difficult for the developers.
    value_error_strings = [ ]

    # Copy/translate all the fields.  field_list is a list of the form
    # fields that we expect from this form.
    for field in field_list :
        if field in request.POST :
            try :
                if not field in converter :
                    # This only happens during development.
                    print("WARNING: field ",field," IS PROPERLY IN FORM, BUT NOT IN etc_web/etc/input_conversion.py")

                elif converter[field] == upload :
                    # Django passes uploaded files in a different field.
                    # Uploaded files are "not in request.POST" according
                    # to the django documentation, but if the user does
                    # not provide a file to be uploaded, the field _does_
                    # occur in request.POST with the value ''
                    value = request.POST[field]
                    if value != '' and value != '-None-' and value != None:
                        raise InputConversionHackDetect(
                            'upload field is not an uploaded file: %s value-%s-'
                            % (field,request.POST[field])
                        )

                elif converter[field] == DISCARD :
                    # DISCARD is the special function for input fields we
                    # should throw away.  If the tests pass with the field
                    # converter set to DISCARD, you can take that field out
                    # of the html for that form.
                    pass

                else :
                    try :
                        # we have to handle both fields that have a plain value, and
                        # fields that have the value encapsulated as a 1-element list.
                        try:
                            value = request.POST[field].pop()
                        except AttributeError:
                            value = request.POST[field]
                        fdict[field] = converter[field](value , field)
                    except :
                        if CFG_DEBUG :
                            sys.stderr.write(
                                "input field failed validation: %s = %s\n"
                                %(field,request.POST[field])
                            )
                            sys.stderr.flush()
                        raise

            # any failure to validate the input value comes here
            except InputConversionError as e :
                value_error_strings.append(str(e))
            except ValueError as e :
                value_error_strings.append(str(e))

            # after processing the field, remove it.  ( If there are extra
            # fields, they will still be in original_inputs at the end. )
            del original_inputs[field]

    # Uploaded files come separately from other fields of the form.
    for field in request.FILES :
        if converter[field] == upload :
            sys.stderr.write("uploaded file: %s\n"%field)
            upload_list.append(field)
        else :
            # bug: instead of saying so, should we just throw "internal error"?
            s = "uploaded file NOT EXPECTED: %s\n"%field
            value_error_strings.append(s)

    sys.stderr.flush()

    # If any original input is left, we have a problem.
    if len(original_inputs) > 0 :
        s = ""
        if CFG_DEBUG :
            for x in original_inputs :
                value_error_strings.append(
                    "UNEXPECTED INPUT during input_conversion: "+str(x)+'='+str(original_inputs[x])+"\n")
                    # Developer note!  If you see this error after adding a new
                    # field to the UI, make sure you have added your new field
                    # name to the correct "form_fields_<inst>_<mode>.py" file !

    # If there were any validation errors, we raise an exception
    # containing the whole list.
    if len(value_error_strings) > 0 :
        raise InputConversionError("\n".join(value_error_strings))

    ###
    # fdict is a dict containing converted/validated values
    # upload_list is a list of the field names for uploaded files
    return ( fdict, upload_list )

#
def upload( value, name ) :
    raise InputConversionHackDetect('this upload() function should not actually get called')

def DISCARD( value, name ) :
    raise InputConversionHackDetect('this DISCARD() function should not actually get called')

##########
#
# Here are various field validators
#

#

def lookup_nice_info(name):
    '''
    look up information to return to the user about the variable to help them understand the error
    '''

    keyword_information={
        'ZodiMult':'Zodiacal light multiplication factor',
        'EarthshineMult': 'Earthshine multiplication factor',
        'fL1_center': 'Line center value in first row',
        'fL1_flux': 'Flux for line center in first row',
        'fL1_fwhm': 'FWHM for line center in first row',
        'fL2_center' : 'Line center value in second row',
        'fL2_flux': 'Flux for line center in second row',
        'fL2_fwhm' : 'FWHM for line center in second row',
        'fL3_center' : 'Line center value in third row',
        'fL3_flux': 'Flux for line center in third row',
        'fL3_fwhm': 'FWHM for line center in third row',
        'extractionRegionUserCircle': 'Point source extraction region at user radius',
        'extractionRegionUserCircle_min': 'Point source extraction region minimum radius',
        'extractionRegionUserCircle_max': 'Point source extraction region maximum radius',
        'extractionRegionExtendedUserCircle': 'Extended source extraction region at user radius',
        'extractionRegionExtendedUserCircle_min': 'Extended source extraction region minimum radius',
        'extractionRegionExtendedUserCircle_max': 'Extended source extraction region maximum radius',
        'extractionRegionVariableCircle': 'Point source extraction region of x% of light',
        'fbbtemp': 'Black Body Temperature',
        'febv': 'E(B-V)',
        'crsplit' : 'CR-Split ("Frames" or "# Indep. Exposures" for WFC3) \n',

        }

    try:
       newname=keyword_information[name]
    except KeyError:
        newname=name
    return newname


re_alnum = re.compile('^[A-Za-z0-9_]*$')

def alnum( value, name ) :
    if re_alnum.match(value) :
        # str() to de-unicode it
        return str(value)
    raise InputConversionError('%s not alphanumeric'%name)

re_al_dot_num = re.compile('^[A-Za-z0-9._]*$')

def al_dot_num( value, name ) :
    if re_al_dot_num.match(value) :
        # str() to de-unicode it
        return str(value)
    raise InputConversionError('%s not valid'%name)


#
re_for_check = { }

def re_check(value, name, regex ) :
    if not regex in re_for_check :
        re_for_check[regex] = re.compile(regex)
    if re_for_check[regex].match(value) :
        # str() to de-unicode it
        return str(value)
    if name == 'fOtherFile':
        raise InputConversionError("Wrong file name: \"" + value +
                        "\"  -  Check spelling of file name in \"Other HST Spectrum\" input field.")
    else:
        raise InputConversionError('%s value %s not matching regex %s'%(name,value,regex))

# WARNING:  overriding builtin functions min() and max() as parameters
def float_range( value, name, min=None, max=None ) :

    nice_name=lookup_nice_info(name)
    try :
        v = float(value.replace(' ',''))
    except :
        raise InputConversionError('%s (%s) is not a floating point (%s)'%(nice_name, value, name))
    if not min is None :
        if v < min :
            raise InputConversionError('%s=%f less than minimum allowed value %f (%s)'%(nice_name, v, min, name))
    if not max is None :
        if v > max :
            raise InputConversionError('%s=%f greater than maximum allowed value %f (%s)'%(nice_name, v, max, name))
    return v

# WARNING:  overriding builtin functions min() and max() as parameters
def int_range( value, name, min=None, max=None ) :
    #get a nice name for the errored key
    nice_name=lookup_nice_info(name)
    try :
        v = int(value)
    except:
        raise InputConversionError('%s (%s) is not an integer (%s)'%(nice_name, value, name))
    if not min is None :
        if v < min :
            raise InputConversionError('%s=%d less than minimum allowed value %d (%s)'%(nice_name, v, min, name))
    if not max is None :
        if v > max :
            raise InputConversionError('%s=%d greater than maximum allowed value %d (%s)'%(nice_name, v, max, name))
    return v

#
def list_int( value, name, l ) :
    try :
        v = int(value)
    except:
        raise InputConversionError('%s is not an integer (%s)'%(lookup_nice_info(name),name))
    if not v in l:
        raise InputConversionError('%s not one of the choices in the list for %s (%s)'%(value,lookup_nice_info(name), name))
    return v

#
def list_str( value, name, valid ) :
    if value in valid :
        # str() to de-unicode it
        return str(value)
    raise InputConversionError('%s not one of the choices in the list for %s'
                     %(value,lookup_nice_info(name)))


# these functions deal with R.A. and Dec. validation
_format_spec = '%s not a valid equatorial coordinate'

def _sex2dec(value):
    try:
        v,m,s = value.split(":")
        v = float(v)
        if v != 0.0:
            signal = v / abs(v)
        else:
            signal = 1.

        result = signal * (abs(v) + float(m)/60. + float(s)/3600.)
        return result

    except ValueError:
        raise InputConversionError(_format_spec % value)

def _check_range(value, value_to_check, min, max):
    # 'value-to_check' is the actual float value to range-check.
    # 'value' is the original input string, used for reporting only.
    if min and value_to_check < min:
        raise InputConversionError(_format_spec % value)
    if max and value_to_check > max:
        raise InputConversionError(_format_spec % value)

#

# A declination looks like
# [+-]00:00:00.00000
#
# degress:minutes:seconds.frac
#
# but etc_engine apparently 1) wants a string, and 2) only allows
# integer seconds
#
re_dec = re.compile('^ *[-+]?[0-9]+:[0-9]+:[0-9]+ *$')

def declination( value, name, min=None, max=None ) :
    # str() to de-unicode it
    value = str(value)
    try:
        # declination in decimal degrees
        value_to_check = float(value)
        _check_range(value, value_to_check, min, max)
        return value

    except ValueError:
        pass

    if not re_dec.match(value) :
        raise InputConversionError(_format_spec % value)

    # declination in sexagesimal degrees
    value_to_check = _sex2dec(value)
    _check_range(value, value_to_check, min, max)
    return value

# A Right Ascension looks like
# 00:00:00.00000
#
# hours:minutes:seconds.frac
#
# but etc_engine apparently 1) wants a string, and 2) only allows
# integer seconds
#
re_ra_1 = re.compile('^ *[0-9]+:[0-9]+:[0-9]+ *$')
# re_ra_1 = re.compile('[0-9]+:[0-9]+:[0-9]+[.]*[0-9]*')
def right_ascension( value, name, min=None, max=None ) :
    # str() to de-unicode it
    value = str(value)
    try:
        # R.A. in decimal degrees
        value_to_check = float(value)
        _check_range(value, value_to_check, min, max)
        return value

    except ValueError:
        pass

    if not re_ra_1.match(value) :
        raise InputConversionError(_format_spec % value)

    # R.A. in hh:mm:ss format
    value_to_check = _sex2dec(value)
    value_to_check *= 360./24.
    _check_range(value, value_to_check, min, max)
    return value


###
#
re_rectangle = re.compile('([0-9]+)|([0-9]+,[0-9]+)')

def rectangle( value, name ) :
    if not re_rectangle :
        raise InputConversionError('%s not a valid rectangle'%name)
    return str(value)

#
default_check = alnum

# This dict contains a list of all valid field names (for all forms)
# and the validation/conversion function to use for each.
#
# Each is on a single very wide line because it makes it easy to keep
# them sorted.  In vi, move to the first line and type "!}./icf" to
# pipe it through the script icf to sort/format the list.  emacs has
# a way to do this too, but I don't remember what it is.  You have
# to use some variant of shell-command-on-region.
#
# It isn't really necessary to give detailed lists here in most cases:
#  - If a value came from a drop-list on the web form, we know it is
#    acceptable for any user who is not trying to circumvent our system. If
#    they _are_, we don't care if they get the wrong answer.
#  - If a value is only ever used as a key to look things up in a dict,
#    it is sufficient to use something simple like alnum.  Values not in the
#    dict will raise an exception anyway, and we don't need to maintain
#    the lists here.
#  - If a value is a number, the engine can't be hurt by it.
#
# Remember the checks here are to guard against being hurt by user input.
# Questions of whether the input values _make_ _sense_ do not belong here.
#

# NOTE - if you add or delete items from here, please consider whether
#        you should also edit anything in etc_web/bit/config ...

converter = {

    'AirglowStandard'                   : alnum,
    'E140H_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1271, 1234, 1307, 1343, 1380, 1416, 1453, 1489, 1526, 1562, 1598]),
    'E140M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1425]),
    'E230H_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1763, 1813, 1863, 1913, 1963, 2013, 2063, 2113, 2163, 2213, 2263, 2313, 2363, 2413, 2463, 2513, 2563, 2613, 2663, 2713, 2762, 2812, 2862, 2912, 2962, 3012]),
    'E230M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1978, 2124, 2269, 2415, 2561, 2707]),
    'EarthshineMag'                     : float_range,
    'EarthshineMult'                    : lambda v,n : float_range(v, n, min=0),
    'EarthshineSpec'                    : alnum,
    'EarthshineStandard'                : alnum,
    'G130M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1055, 1096, 1222, 1291, 1300, 1309, 1318, 1327, ] ),
    'G140L_CentralWavelength'           : lambda v,n : list_int( v,n, [ 800, 1105, 1280, ]),
    'G140M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1173, 1218, 1222, 1272, 1321, 1371, 1387, 1400, 1420, 1470, 1518, 1540, 1550, 1567, 1616, 1640, 1665, 1714]),
    'G160M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1533, 1577, 1589, 1600, 1611, 1623, ]),
    'G185M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1786, 1817, 1835, 1850, 1864, 1882, 1890, 1900, 1913, 1921, 1941, 1953, 1971, 1986, 2010, ]),
    'G225M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 2186, 2217, 2233, 2250, 2268, 2283, 2306, 2325, 2339, 2357, 2373, 2390, 2410, ] ),
    'G230L_CentralWavelength'           : lambda v,n : list_int( v,n, [ 2635, 2950, 3000, 3360, ] ),
    'G230MB_CentralWavelength'          : lambda v,n : list_int( v,n, [ 1713, 1854, 1995, 2135, 2276, 2416, 2557, 2697, 2794, 2836, 2976, 3115] ),
    'G230M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 1687, 1769, 1851, 1884, 1933, 2014, 2095, 2176, 2257, 2338, 2419, 2499, 2579, 2600, 2659, 2739, 2800, 2818, 2828, 2898, 2977, 3055]),
    'G285M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 2617, 2637, 2657, 2676, 2695, 2709, 2719, 2739, 2850, 2952, 2979, 2996, 3018, 3035, 3057, 3074, 3094, ]),
    'G430M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 3165, 3305, 3423, 3680, 3843, 3936, 4194, 4451, 4706, 4781, 4961, 5093, 5216, 5471] ),
    'G750M_CentralWavelength'           : lambda v,n : list_int( v,n, [ 5734, 6094, 6252, 6581, 6768, 7283, 7795, 8311, 8561, 8825, 9286, 9336, 9806, 9851] ),
    'HeliumStandard'                    : lambda v,n : list_str(v,n, ['None','Average','High','Very High']),
    'Mirror'                            : lambda v,n : list_str(v,n, ['mirrora', 'mirrorb' ]),
    'SNR'                               : float_range,
    'Stripe'                            : lambda v,n : list_str(v, n, ['A', 'B', 'C'] ),
    'Time'                              : lambda v,n : float_range(v, n, min=0),
    'scRate'                            : lambda v,n : float_range(v, n, min=0),
    'scLengthFromRate'                  : lambda v,n : float_range(v, n, min=0),
    'scLengthFromTime'                  : lambda v,n : float_range(v, n, min=0),
    'ZodiDate'                          : lambda v,n : int_range(v,n, min=1990),
    'ZodiDay'                           : lambda v, n: int_range(v, n, min=1, max=31),
    'ZodiDec'                           : lambda v, n: declination(v, n, min=-90.0, max=90.0),
    'ZodiMag'                           : float_range,
    'ZodiMonth'                         : lambda v, n: int_range(v, n, min=1, max=12),
    'ZodiMult'                          : lambda v,n : float_range(v, n, min=0),
    'ZodiRA'                            : lambda v, n: right_ascension(v, n, min=0.0, max=360.0),
    'ZodiSpec'                          : alnum,
    'ZodiStandard'                      : alnum,
    'ZodiSun'                           : lambda v,n : float_range(v, n, min=0.0, max=360.0),
    'ZodiSunAttribute'                  : alnum,
    'binx'                              : lambda v,n : list_int(v, n, [ 1,2,4 ]),
    'biny'                              : lambda v,n : list_int(v, n, [ 1,2,4 ]),
    'ccdMode'                           : lambda v,n : list_str(v, n, [ 'ACQ', 'ACQ/PEAK' ]),
    'ccdaperture0'                      : lambda v,n : list_str(v, n, ['0.2X0.2','52X0.05','52X0.1','52X0.2','52X0.5','52X2','50CCD','F28X50LP','F25ND3','F25ND5','F28X50OII','F28X50OIII','F25NDQ1','F25NDQ2','F25NDQ3','F25NDQ4','0.2X0.05ND','0.3X0.05ND','0.1X0.03','0.2X0.06','0.2X0.09','31X0.05NDA','31X0.05NDB','31X0.05NDC']),
    'ccdapertureACQ'                    : al_dot_num,
    'ccdapertureACQPEAK'                : al_dot_num,
    'ccddarklevel'                      : lambda v,n : list_str(v, n, [ 'high', 'medium', 'low' ] ),
    'cosaperture0'                      : lambda v,n : list_str(v, n, [ 'PSA', 'BOA' ] ),
    'crsplit'                           : lambda v, n: int_range(v, n, min=1),
    'detector'                          : alnum,
    'disperser'                         : alnum,
    'e140haperture0'                    : lambda v,n : list_str(v, n, ['0.2X0.09','0.2X0.2','6X0.2','0.1X0.03','0.2X0.05ND','0.3X0.05ND', '52X0.05','25MAMA', 'F25QTZ','F25SRF2', 'F25ND3', 'F25ND5','31X0.05NDA','31X0.05NDB','31X0.05NDC']),
    'e140maperture0'                    : lambda v,n : list_str(v, n, ['0.2X0.06','0.2X0.2','6X0.2','0.1X0.03','0.2X0.05ND','0.3X0.05ND', '52X0.05','25MAMA', 'F25QTZ','F25SRF2', 'F25ND3', 'F25ND5','31X0.05NDA','31X0.05NDB','31X0.05NDC']),
    'e230haperture0'                    : lambda v,n : list_str(v, n, ['0.2X0.09','0.2X0.2','6X0.2','0.1X0.03','0.1X0.09','0.1X0.2','0.2X0.05ND','0.3X0.05ND', '52X0.05','25MAMA', 'F25QTZ','F25SRF2', 'F25MGII', 'F25ND3', 'F25ND5','31X0.05NDA','31X0.05NDB','31X0.05NDC']),
    'e230maperture0'                    : lambda v,n : list_str(v, n, ['0.2X0.06','0.2X0.2','6X0.2','0.1X0.03','0.2X0.05ND','0.3X0.05ND', '52X0.05','25MAMA', 'F25QTZ','F25SRF2', 'F25MGII', 'F25ND3', 'F25ND5','31X0.05NDA','31X0.05NDB','31X0.05NDC']),
    'extractionRegionCircle'            : lambda v,n : float( list_str( v, n, ["0.1", "0.2", "0.3", "0.4", "0.5", "0.6", "0.8", "1.0", "1.5", "2.0", 0.4,] )),
    'extractionRegionExtendedCircle'    : lambda v,n : float( list_str( v, n, ["0.1", "0.2", "0.3", "0.4", "0.5", "0.6", "0.8", "1.0", "1.5", "2.0", 0.4,] )),
    'extractionRegionExtendedSquare'    : lambda v,n : int_range( v, n, 0, 200 ),
    'extractionRegionExtendedUserCircle': lambda v,n : float_range(v, n, min=0),
    'extractionRegionExtendedUserCircle_min': float_range,
    'extractionRegionExtendedUserCircle_max': float_range,
    'extractionRegionRectangle'         : rectangle,
    'extractionRegionScHeight'          : alnum,
    'extractionRegionScWidth'           : alnum,
    'extractionRegionSquare'            : lambda v,n : list_int( v, n, [ 1, 2, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 51, 101, ] ),
    'extractionRegionUserCircle'        : lambda v,n : float_range(v, n, min=0),
    'extractionRegionUserCircle_min'    : float_range,
    'extractionRegionUserCircle_max'    : float_range,
    'extractionRegionVariableCircle'    : lambda v,n : float_range(v, n, min=0, max=100.),
    'fband'                             : alnum,
    'fbandpass'                         : alnum,
    'fIndex'                            : float_range,
    'fIsLambda'                         : lambda v,n : list_str(v,n,['true', 'false'] ),
    'fL1_center'                        : lambda v,n : float_range(v, n, min=0),
    'fL1_flux'                          : lambda v,n : float_range(v, n, min=0),
    'fL1_fwhm'                          : lambda v,n : float_range(v, n, min=0),
    'fL2_center'                        : lambda v,n : float_range(v, n, min=0),
    'fL2_flux'                          : lambda v,n : float_range(v, n, min=0),
    'fL2_fwhm'                          : lambda v,n : float_range(v, n, min=0),
    'fL3_center'                        : lambda v,n : float_range(v, n, min=0),
    'fL3_flux'                          : lambda v,n : float_range(v, n, min=0),
    'fL3_fwhm'                          : lambda v,n : float_range(v, n, min=0),
    'fOtherFile'                        : lambda v,n : re_check(v,n, '^[$A-Za-z0-9._]*$'), # file names in cdbs in IRAF format
    'fRedshift'                         : lambda v,n : float_range(v, n, min=0),
    'fSpectrumCK'                       : alnum,
    'fSpectrumPickles'                  : alnum,
    'fStellar'                          : al_dot_num,
    'fUploadFile'                       : upload,
    'fbbtemp'                           : lambda v,n : float_range(v, n, min=0),
    'fbpgsfile'                         : alnum,
    'fcalfile'                          : lambda v,n : re_check(v,n, '^[A-Za-z0-9._+-]*$'),
    'fphoenixfile'                      : lambda v,n : re_check(v,n, '^[A-Za-z0-9._+-]*$'),
    'fdiameter'                         : float_range,
    'febmvtype'                         : lambda v,n : list_str(v,n,[ "mwavg", "mwdense", "mwrv21","mwrv4","lmcavg", "lmc30dor", "smcbar", "xgalsb"]),
    'febv'                              : lambda v,n : float_range(v, n, min=0),
    'fextinctiontype'                   : lambda v,n : list_str(v,n,[ "before", "after" ]),
    'rn_flux_lambda'                    : float_range,
    'rn_flux_lambda_units'              : lambda v,n : list_str(v, n,['flam','jy','abmag']),
    'rn_flux_bandpass'                    : float_range,
    'rn_flux_bandpass_units'              : lambda v,n : list_str(v, n,['flam','jy','abmag', 'vegamag']),
    'rn_lambda'                         : float_range,
    'rn_lambda_units'                   : lambda v,n : list_str(v, n,['angstroms','nanometers','microns']),
    'fftype'                            : al_dot_num,
    'fftype_filters'                    : al_dot_num,
    'fnonstellar'                       : lambda v,n : list_str(v,n, [
        'Gliese 229B',
        'Gliese 752B',
        'Gliese 411',
        'Gliese 406',
        'Orion',
        'Orion2',
        'PN',
        'PN Extended',
        'QSO',
        'QSO2',
        'QSO SDSS',
        'QSO IRTF',
        'QSO COS ',
        'Elliptical',
        'Elliptical2',
        'El B2004a' ,
        'El CWW FUV',
        'NGC1068',
        'Im B2004a',
        'Spiral',
        'S0 FUV CWW',
        'Sa FUV CWW',
        'Sbc B2004a',
        'Sbc CWW',
        'Scd B2004a',
        'SSP 25myr z008',
        'SSP 5myr z008',
        'Tau06 z02 1015 190',
        'Tau06 z02 2500 000',
        'Tau06 z02 4500 000',
        'Tau06 z02 5000 000',
        'Tau06 z02 6000 000',
        'Tau06 z02 8000 000',
        'Tau06 z02 12000 000',
        'Sb1 Kinney FUV',
        'Sb2 B2004a',
        'Sb2 Kinney FUV',
        'Sb3 B2004a',
        'Sb3 Kinney FUV',
        'Sb4 Kinney FUV',
        'Sb5 Kinney FUV',
        'Sb6 Kinney FUV',
        'ULIRG05189-2524',
        'ULIRG12112+0305',
        'ULIRG14348+1447',
        'ULIRG15250+3609',
        'JWIRAS22491-1808',
        'ARP220',
        'SB_M82',
        'MRK_1014',
        'MRK_231',
        'MRK_273',
        'MRK_463',
        'NGC6240',
        'UGC5101'
        ]),
    'fsorigin'                          : lambda v,n : list_str(v, n, [ 'SpectrumUpload', 'SpectrumHstOther', 'SpectrumCK', 'SpectrumPickles', 'SpectrumKurucz', 'SpectrumStellar', 'SpectrumHST', 'SpectrumPhoenix', 'SpectrumNonStellar', 'SpectrumBlackBody', 'SpectrumPowerLaw', 'SpectrumFlat', 'SpectrumEmpty', ] ),
    'fsourceType'                       : lambda v,n : list_str(v, n, [ 'extended', 'point' ] ),
    'filter.ubvri'                      : lambda v,n : list_str(v,n, ["Johnson/U", "Johnson/B", "Johnson/V", "Johnson/R", "Johnson/I", "Johnson/J", "Johnson/K", "Bessell/H", "Bessell/K", "Bessell/J","Cousins/R", "Cousins/I"]),
    'filter.sloan'                      : lambda v,n : list_str(v,n, ["Galex/FUV", "Galex/NUV", "Sloan/U", "Sloan/G", "Sloan/R", "Sloan/I", "Sloan/Z"]),
    'filter.nicmos'                     : lambda v,n : list_str(v,n, ["NICMOS/F110W", "NICMOS/F160W"]),
    'filter.acs'                        : lambda v,n : list_str(v,n, ["ACS/F435W", "ACS/F475W", "ACS/F555W", "ACS/F606W", "ACS/F625W", "ACS/F775W", "ACS/F814W", "ACS/F850LP"]),
    'filter.wfc3UVIS'                   : lambda v,n : list_str(v,n, ["WFC3/UVIS/F218W", "WFC3/UVIS/F200LP", "WFC3/UVIS/F225W", "WFC3/UVIS/F275W", "WFC3/UVIS/F300X", "WFC3/UVIS/F336W", "WFC3/UVIS/F350LP", "WFC3/UVIS/F390W", "WFC3/UVIS/F438W", "WFC3/UVIS/F475X","WFC3/UVIS/F600LP"]),
    'filter.wfc3IR'                     : lambda v,n : list_str(v,n, ["WFC3/IR/F098M", "WFC3/IR/F105W", "WFC3/IR/F110W", "WFC3/IR/F125W", "WFC3/IR/F140W", "WFC3/IR/F160W"]),
    'fuvmamaaperture0'                  : lambda v,n : list_str(v, n, ['0.2X0.2','52X0.05','52X0.1','52X0.2','52X0.5','52X2','25MAMA','F25NDQ1','F25NDQ2','F25NDQ3','F25NDQ4','F25ND3','F25ND5','F25QTZ','F25SRF2','F25LYA','31X0.05NDA','31X0.05NDB','31X0.05NDC']),
    'gain'                              : float_range,
    'hrcfilt0'                          : lambda v,n : alnum(v,n),
    'hrcfilt1'                          : lambda v,n : alnum(v,n),
    'instrument'                        : alnum,
    'irfilt0'                           : alnum,
    'cosAcqMode'                        : lambda v,n : list_str(v, n, [ 'image', 'search' ] ),
    'fuvglowregion'                     : lambda v,n : list_str(v, n, [ 'high', 'medium', 'low', 'none' ] ),
    'nuvMode'                           : lambda v,n : list_str(v, n, ['ACQ', 'ACQ/PEAKXD','ACCUM'] ),
    'nuvmamaaperture0'                  : lambda v,n : list_str(v, n, ['0.2X0.2','52X0.05','52X0.1','52X0.2','52X0.5','52X2','25MAMA','F25NDQ1','F25NDQ2','F25NDQ3','F25NDQ4','F25ND3','F25ND5','F25QTZ','F25SRF2', 'F25MGII', 'F25CN270','F25CN182', 'F25CIII','31X0.05NDA','31X0.05NDB','31X0.05NDC']),
    'obswave'                           : lambda v,n : float_range(v, n, min=0, max=20000),
    'post_flash_acs'                    : lambda v, n: int_range(v, n, min=0, max=5733),
    'post_flash_wfc3'                   : lambda v, n: int_range(v, n, min=0, max=25),
    'prismaperture0'                    : lambda v,n : list_str(v, n, ['52X0.05', '52X0.1','52X0.2','52X0.5','52X2','F25QTZ','F25SRF2','F25MGII','F25NDQ1','F25NDQ2','F25NDQ3','F25NDQ4','F25ND3','F25ND5','25MAMA']),
    'sbcfilt0'                          : lambda v,n : list_str(v, n, [ 'F122M','F115LP','F125LP','F140LP','F150LP','F165LP'] ),
    'science_mode'                      : alnum,
    'simmode'                           : lambda v,n : list_str(v, n, [ 'Time', 'SNR', 'SNR_from_rate' ]),
    'wavelengthUnits'                   : alnum,
    'wfc3_filter_m'                     : alnum,
    'wfc3_filter_n'                     : alnum,
    'wfc3_filter_q'                     : alnum,
    'wfc3_filter_type'                  : lambda v,n : list_str(v, n, [ 'medium','wide','narrow','quad']),
    'wfc3_filter_w'                     : alnum,
    'wfcfilt0'                          : lambda v,n : alnum(v,n),
    'wfcfilt1'                          : lambda v,n : alnum(v,n),
    'xRegionExtendedType'               : lambda v,n : list_str(v, n, ['default','Default', 'Circle', 'Square', 'Rectangle', 'UserCircle'] ),
    'xRegionType'                       : lambda v,n : list_str(v, n, [ 'default', 'Default', 'Square', 'Circle', 'UserCircle', 'VariableCircle', 'Rectangle', 'scRectangle' ] ),
}

#
# All of this is a consistency check that happens in developer systems.
# - are there fields in the form that are not listed in converter[] ?
# - are there fields in converter[] that are not listed in any form ?
#

if CFG_DEBUG :
    check_fields_missing = { }
    extra_fields = converter.copy()

    def check_fields(b) :
        for x in b.field_list :
            if not x in converter :
                if x in check_fields_missing:
                    check_fields_missing[x].append( b.__file__ )
                else :
                    check_fields_missing[x] = [ b.__file__ ]
            if x in extra_fields :
                del extra_fields[x]

    import pandeia.engine.helpers.bit.instruments.hst.acs.web.form_fields_acs_imaging as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.acs.web.form_fields_acs_rampfilter as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.acs.web.form_fields_acs_spectroscopic as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.cos.web.form_fields_cos_imaging as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.cos.web.form_fields_cos_spectroscopic as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.cos.web.form_fields_cos_spectroscopicacq as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.cos.web.form_fields_cos_targetacquisition as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.stis.web.form_fields_stis_imaging as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.stis.web.form_fields_stis_spectroscopic as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.stis.web.form_fields_stis_targetacquisition as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.wfc3ir.web.form_fields_wfc3ir_imaging as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.wfc3ir.web.form_fields_wfc3ir_scimaging as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.wfc3ir.web.form_fields_wfc3ir_spectroscopic as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.wfc3ir.web.form_fields_wfc3ir_scspectroscopic as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.wfc3uvis.web.form_fields_wfc3uvis_imaging as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.wfc3uvis.web.form_fields_wfc3uvis_scimaging as a
    check_fields(a)
    import pandeia.engine.helpers.bit.instruments.hst.wfc3uvis.web.form_fields_wfc3uvis_spectroscopic as a
    check_fields(a)

    if len(check_fields_missing) > 0 :
        sys.stderr.write(
            "\n\nMISSING FIELDS (required by some form) IN INPUT VALIDATOR: %s\n"
            %__file__
        )
        for X in check_fields_missing :
            sys.stderr.write( "    %s\n" % X )
            for y in check_fields_missing[X] :
                sys.stderr.write("            %s\n" % os.path.basename(y))
        sys.stderr.write('\n')
        sys.stderr.flush()

    if len(extra_fields) > 0 :
        sys.stderr.write(
            "\n\nEXTRA FIELDS (not expected by any form) IN INPUT VALIDATOR: %s\n"
            % __file__
        )
        for X in extra_fields :
            sys.stderr.write("    %s\n" % X)
        sys.stderr.write("\n")
        sys.stderr.flush()

# =========================================================================

PLOT_TYPES = [
    "Observed Target Spectrum",
    "Input Target Spectrum",
    "Signal-to-noise",
    "Throughput",
    "Total Counts",
    "Re-plot"
    ]

LEGEND_LOCATIONS = [
    "best",
    "upper right",
    "center right",
    "lower right",
    "upper left",
    "center left",
    "lower left",
    "upper center",
    "center",
    "lower center",
]

INT_RE = re.compile(r"\d+")
def chart_pixels(pstr):
    """Convert to integer and validate a chart X or Y pixel dimension."""
    if INT_RE.match(pstr):
        pix = int(pstr)
        if 350 <= pix <= 3000:
            return pix
    raise InputConversionError("size must be a decimal number in the range 350..3000")

FLOAT_RE = re.compile(r"^([+/-]?((([0-9]+(\.)?)|([0-9]*\.[0-9]+))([eE][+\-]?[0-9]+)?))$")
def valid_float(fstr):
    """Convert suitable strings into floats,  exclude expressions."""
    if FLOAT_RE.match(fstr):
        return float(fstr)
    raise InputConversionError("badly formatted real number " + repr(str(fstr)))

#
# These are all the inputs that may be given to a plot request, along
# with the validator to use for it.
#
PLOT_INPUTS =     {   # Map form variables to validator kinds
        "LogX" : "checkbox",
        "LogY" : "checkbox",
        "ChartXMin" : valid_float,
        "ChartXMax" : valid_float,
        "ChartYMin" : valid_float,
        "ChartYMax" : valid_float,
        "ChartXSize" : chart_pixels,
        "ChartYSize" : chart_pixels,
        "FullRange" : "checkbox",
        "PlotType" : PLOT_TYPES,
        "PlotTypeVal": PLOT_TYPES,
        "LegendLocation":LEGEND_LOCATIONS,
      }

PLOT_ERR_MAP = {   # Map form variables to plot page oriented text
        "ChartXMin" : "Axis Min, Wavelength",
        "ChartYMin" : "Axis Min, Y",
        "ChartXMax" : "Axis Max, Wavelength",
        "ChartYMax" : "Axis Max, Y",
        "ChartXSize" : "Plot Size, Wavelength",
        "ChartYSize" : "Plot Size, Y",
        "LegendLocation" : "Legend Location",
    }

def plot_input_conversion(request):
    """Fetches and validates plot form POST variables from `request`
    returning clean form variable dictionary `fdict`.
    """
    # We collect all the valid inputs in fdict and then return it.
    # Invalid fields are just ignored because the are not listed in PLOT_INPUTS.
    fdict = {}
    for var, validator in list(PLOT_INPUTS.items()):
        if var in request.POST:
            value = request.POST[var]
            if value == "":
                continue
            if isinstance(validator, list):
                if value in validator:
                    fdict[var] = value
                else:
                    raise InputConversionError("Unexpected value for " + repr(PLOT_ERR_MAP[var]) + " not one of: " + str(validator))
            elif validator == "checkbox":
                fdict[var] = "checked"
            else:
                try:
                    fdict[var] = validator(value)
                except InputConversionError as e:
                    raise InputConversionError("Invalid " + repr(PLOT_ERR_MAP[var]) + " : " + str(e))
                except:
                    raise InputConversionError("Invalid input for " + repr(PLOT_ERR_MAP[var]) + " = " + repr(str(value)))
    return fdict



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import copy
from warnings import warn
import numpy as np
import synphot as syn
from synphot.models import Empirical1D, ConstFlux1D
import astropy.units as u

from . import config as cf
from .constants import PANDEIA_WAVEUNITS, PANDEIA_FLUXUNITS, PANDEIA_ATOL, PANDEIA_RTOL
from .config import DefaultConfig
from .io_utils import ref_data_column
from .utils import spectrum_resample
from .custom_exceptions import EngineInputError, DataError, SynphotError, WavesetMismatch

default_refdata_directory = cf.default_refdata_directory


class Background(cf.DefaultConfig):

    """
    This class provides the background flux.  The background can be computed from the engine's
    notional reference data or passed as a spectrum of a form [wave, sb] where wave is an array of
    wavelengths in microns and sb is an array of surface brightnesses in MJy/sr.

    Parameters
    ----------
    observation: Observation instance
        Contains needed instrument and background configuration data.
    config: dict
        Extra configuration information
    webapp: bool
        Enable strict API checking (not yet implemented here)

    Attributes
    ----------
    ref_dir: string
        Directory containing data for notional background model
    bg_level: string or list-like or array-like
        Input background level. Can be string for configuring notional model or lists/arrays for user-input background.
    instrument: Instrument instance
        Instrument configuration data is required to determine pixel scale and telescope/instrument contributions
        to the background model.
    aperture_pars: dict
        Instrument configuration data containing pixel scale required to calculate per-pixel background flux
    bg_spec: synphot.SourceSpectrum
        Original full-length, full-resolution background spectrum with wavelengths in microns and surface brightness in MJy/sr.
    bg: synphot.SourceSpectrum
        Current, possibly resampled background spectrum with wavelengths in microns and surface brightness in MJy/sr.

    Properties
    ----------
    wave: 1D np.ndarray
        Binned/trimmed wavelength set
    MJy_sr: 1D np.ndarray
        Binned/trimmed background spectrum in units of MJy/sr
    mjy_pix: 1D np.ndarray
        Convert self.MJy_sr to units of mJy (milli-janskys) per pixel
    flambda_pix: 1D np.ndarray
        Convert self.MJy_sr to units of F_lambda per pixel, photons/cm^2/s/micron/pixel.
    """

    def __init__(self, observation, config={}, webapp=False, **kwargs):
        self.webapp = webapp
        self.telescope = observation.instrument.telescope.tel_name

        DefaultConfig.__init__(self, config=config, webapp=webapp, **kwargs)

        self.ref_dir = os.path.join(default_refdata_directory, self.telescope, 'background')
        self.bg_location = observation.background
        self.instrument = observation.instrument
        self.aperture_pars = self.instrument.get_aperture_pars()
        # if the background spectrum is provided, the 0th index is wavelength and the 1st is the surface brightness
        if isinstance(self.bg_location, (list, tuple, np.ndarray)):
            try:
                # wavelength set for background that will get resampled to match wavelength set for a scene
                wave = np.array(self.bg_location[0], dtype=np.float64) << PANDEIA_WAVEUNITS
                # background spectrum in standard units of MJy/sr. use properties mjy_pix and flambda_pix to convert units.
                MJy_sr = np.array(self.bg_location[1], dtype=np.float64) << u.MJy
                # we store the original bg spectrum in a synphot spectrum so that resampling operations
                # are done on the original data.
                self.bg_spec = syn.spectrum.SourceSpectrum(Empirical1D, points=wave, lookup_table=MJy_sr)
            except Exception as e:
                msg = "Malformed input background spectrum: %s (%s)" % (repr(self.bg_location), e)
                raise EngineInputError(value=msg)
        elif isinstance(self.bg_location, str):
            if self.bg_location == "none":
                pass
            else:
                self.bg_level = observation.background_level
            self.get_canned_background()
        else:
            msg = "Must provide a valid background spectrum or background signifier string: %s" % repr(self.bg_location)
            raise EngineInputError(value=msg)
        self.bg = copy.deepcopy(self.bg_spec)

    def _get_config(self):
        """
        Read default configuration from JSON
        defaults are grouped based on telescope, get the background for the specified telescope

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # make use of the base class method to get the config
        config = super()._get_config()[self.telescope]
        config["bg_location_default"] = config["bg_location"]
        config["bg_level_default"] = config["bg_level"]
        if self.webapp:
            config.pop("bg_location", None)
            config.pop("bg_level", None)
        return config

    def resample(self, wavelengths):
        """
        Re-bin background onto a new set of wavelengths.

        This method rewrites bg, always starting from the original unmodified bg_spec

        Parameters
        ----------
        wavelengths: 1D numpy array
            Array of new wavelengths to sample spectrum onto
        """
        # Yes, spectrum_resample expects input in mJy and outputs in mJy, but given that there's a constant
        # 1e9 factor also in fnu space between them, the results should be valid as MJy.
        MJy_sr = spectrum_resample(self.bg_spec(self.bg_spec.waveset, u.MJy).to_value(u.MJy), \
                                   self.bg_spec.waveset.to_value(PANDEIA_WAVEUNITS), wavelengths)
        self.bg = syn.spectrum.SourceSpectrum(Empirical1D, points=wavelengths << PANDEIA_WAVEUNITS, lookup_table=MJy_sr << u.MJy)

    def get_canned_background(self):
        """
        Read reference data for pandeia's canned background models and configure self.wave, self.bg_location,
        self.bg_level, and self.bg_spec to use it.
        """

        if self.bg_location == 'none' or self.bg_level == 'none':
            bg_file = os.path.join(default_refdata_directory, self.telescope, "background", "{0}_{1}.fits".format(self.bg_location_default,self.bg_level_default))
            bg_wave = ref_data_column(bg_file, colname='wavelength') << PANDEIA_WAVEUNITS
            bg_sb = ref_data_column(bg_file, colname='background') * 0 << u.MJy #/u.sr, but that's not valid in spectra
            try:
                # Using ConstFlux1D would work, but then the SourceSpectrum has no waveset, and the rest of the functions expect one.
                tot_bg = syn.spectrum.SourceSpectrum(Empirical1D, points=bg_wave, lookup_table=bg_sb)
            except Exception as e:
                msg = "Problem creating background spectrum with Synphot."
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise SynphotError(value=msg)
        else:
            bg_file = os.path.join(default_refdata_directory, self.telescope, "background", "{0}_{1}.fits".format(self.bg_location,self.bg_level))
            try:
                bg_wave = ref_data_column(bg_file, colname='wavelength') << PANDEIA_WAVEUNITS
                bg = ref_data_column(bg_file, colname='background') << u.MJy #/u.sr, but that's not valid in spectra
                tot_bg = syn.spectrum.SourceSpectrum(Empirical1D, points=bg_wave, lookup_table=bg)
            except Exception as e:
                msg = "Problem loading background model. "
                if self.webapp:
                    msg += "(%s)" % type(e)
                else:
                    msg += repr(e)
                raise DataError(value=msg)

        self.bg_spec = tot_bg

    @property
    def wave(self):
        return self.bg.waveset.to_value(PANDEIA_WAVEUNITS)

    @property
    def MJy_sr(self):
        return self.bg(self.bg.waveset, flux_unit=u.MJy).to_value(u.MJy)


    @property
    def mjy_pix(self):
        """
        Convert background surface brightness in MJy/sr to F_nu per pixel in
        mJy/pixel (self.wave assumed to be in microns)

        Returns
        -------
        i_pix: 1D np.ndarray
            Flux per pixel in mJy/pixel
        """
        i_nu = self.bg(self.bg.waveset, flux_unit=PANDEIA_FLUXUNITS).to_value(PANDEIA_FLUXUNITS) * (u.arcsec * u.arcsec).to(u.sr)

        i_nu_pix = i_nu * self.aperture_pars['plate_scale'][0] * self.aperture_pars['plate_scale'][1]
        return i_nu_pix

    def trim(self, wmin, wmax):
        """
        Trim spectrum to wavelength range specified by wmin and wmax.
        This function does not use Astropy Quantities (except at the beginning and end)
        Like resample, it resamples off of the original background wavelength array

        Parameters
        ----------
        wmin: float
            Minimum wavelength (microns)
        wmax: float
            Maximum wavelength (microns)
        """
        wave = self.bg_spec.waveset.to_value(PANDEIA_WAVEUNITS)
        # Convert to MJy (synphot), output in units of MJy (astropy).
        flux = self.bg_spec(self.bg_spec.waveset, flux_unit=u.MJy).to_value(u.MJy)
        valid_subs = np.where((wave >= wmin) & (wave <= wmax))
        if valid_subs is None or len(valid_subs[0]) == 0:
            # corner case: We have a spectrum that covers the wavelength range 
            # of the instrument setup, but it's so sparsely specified that no 
            # points fall within the range.
            if (wave[0] <= wmin) and (wave[-1] >= wmax):
                wave = np.asarray([wmin,wmax])
                valid_subs = slice(2)
                flux = self.bg_spec(wave << PANDEIA_WAVEUNITS, flux_unit=u.MJy).to_value(u.MJy)
            else:
                msg = "Spectrum for source does not have any overlap with instrument wavelength range."
                raise WavesetMismatch(value=msg)

        trim_wave = []
        # make sure trimmed wavelengths include wmin and wmax
        if not np.isclose(wave[valid_subs][0], wmin, atol=PANDEIA_ATOL, rtol=PANDEIA_RTOL):
            trim_wave.append(wmin)

        trim_wave.extend(wave[valid_subs])

        if not np.isclose(wave[valid_subs][-1], wmax, atol=PANDEIA_ATOL, rtol=PANDEIA_RTOL):
            trim_wave.append(wmax)

        trim_wave = np.array(trim_wave)

        # this will use synphot to properly sample the endpoints
        # It requires unitless data
        trim_flux = spectrum_resample(flux, wave, trim_wave)

        self.nw = wave.size
        # re-add units
        self.bg = syn.spectrum.SourceSpectrum(Empirical1D, points=trim_wave << PANDEIA_WAVEUNITS, lookup_table=trim_flux << u.MJy)

    @property
    def flambda_pix(self):
        """
        Convert background surface brightness in MJy/sr to F_lambda per pixel in
        photons/cm^2/s/micron/pixel (self.wave assumed to be in microns)

        Returns
        -------
        i_pix: 1D np.ndarray
            Flux per pixel in photons/s/cm^2/pixel
        """
        i_lambda = 1.50916e9 / self.wave * self.bg(self.bg.waveset, flux_unit=syn.units.FLAM).value << (u.arcsec * u.arcsec).to(u.sr)
        i_lambda_pix = i_lambda * self.aperture_pars['plate_scale'][0] * self.aperture_pars['plate_scale'][1]
        return i_lambda_pix



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import json
import pkg_resources

import numpy as np

from . import config
from .io_utils import read_json
from pandeia.engine.source import Source, Line, load_empty_source
from pandeia.engine.custom_exceptions import EngineInputError
from pandeia.engine.normalization import NormalizationFactory
from pandeia.engine.sed import SEDFactory
from pandeia.engine.utils import apply_default_only
from pandeia.engine.config import DefaultConfig

default_refdata_directory = config.default_refdata_directory

def strip_meta(d):
    """
    recursively strip any 'meta' tags from a dict, d. this function modifies the dict in place.
    """
    if isinstance(d, dict):
        if 'meta' in d:
            d.pop('meta')
        for k in d:
            strip_meta(d[k])


def get_telescope_config(telescope):
    """
    Get the telescope configuration to tell us instruments and modes.
    Must have the pandeia_refdata env variable set and pointing to the reference data

    Parameters
    ----------
    telescope: string
        Telescope identifier string. Currently supports jwst, roman, and hst.

    Returns
    -------
    config: dict
        Telescope configuration data
    """
    config_file = os.path.join(default_refdata_directory, telescope, 'telescope', 'config.json')
    config = read_json(config_file, raise_except=True)
    strip_meta(config)
    return config


def get_instrument_config(telescope, instrument):
    """
    Get the instrument configuration.
    Must have the pandeia_refdata env variable set and pointing to the reference data

    Parameters
    ----------
    telescope: string
        Telescope identifier string. Currently supports jwst, roman, and hst.
    instrument: string
        Instrument identifier string.

    Returns
    -------
    config: dict
        Instrument configuration data
    """
    config_file = os.path.join(default_refdata_directory, telescope, instrument, 'config.json')
    config = read_json(config_file, raise_except=True)
    strip_meta(config)
    return config


def build_default_source(telescope, geometry="point"):
    """
    Build and return a default Source in engine API dict format

    Returns
    -------
    src: dict
        Source configuration data in engine API format
    """
    src = Source(telescope, shape={"geometry": geometry}, webapp=False).as_dict()

    src['spectrum']['normalization'] = NormalizationFactory(config=src['spectrum']['normalization'], webapp=False).as_dict()
    src['spectrum']['sed'] = SEDFactory(config=src['spectrum']['sed'], webapp=False).as_dict()
    return src


def build_default_line():
    """
    Build and return a default Line in engine API dict format

    Returns
    -------
    line: dict
        Line configuration data in engine API format
    """
    line = Line(webapp=False).as_dict()
    return line


def build_default_scene(telescope):
    """
    Build and return a default scene which consists of a single default Source

    Returns
    -------
    scene: list
        Single element list containing a source as built by build_default_source()
    """
    scene = [build_default_source(telescope)]
    return scene


def build_empty_scene():
    """
    Build and return an empty scene. Because of the way ConvolvedSceneCubes and AstroSpectrums are created,
    there must be some concept of a spectrum and thus a source contained within a scene. The (admittedly
    hacky) workaround is to build a default scene with a flat-spectrum point source and set its flux to 0.
    This will provide the framework necessary to build things from this scene using the existing code, but not
    add any actual flux. The primary motivation for this is the BNC which needs to build a ConvolvedSceneCube
    that only contains background signal.

    As for the scene itself, we have to define a complete source here; the idea of an empty scene depending on
    the telescope is perhaps a bit too much, and we need to define the complete normalization here anyway (for
    some units, e.g. Vegamags, 0 is bright, not faint.)

    As of 2022.0606, this definition is equivalent to the default JWST/Roman source with flux set to zero. It
    must be kept up-to-date with our source API.

    Returns
    -------
    scene: list
        Single element list containing a zero-flux source
    """
    src = load_empty_source()
    src['spectrum']['normalization'] = NormalizationFactory(config=src['spectrum']['normalization'], webapp=False).as_dict()
    src['spectrum']['sed'] = SEDFactory(config=src['spectrum']['sed'], webapp=False).as_dict()
    src['spectrum']['normalization']['norm_flux'] = 0.0
    return [src]

def build_default_calc(telescope, instrument, mode, **kwargs):
    """
    Build a default calculation given an telescope, instrument, and mode

    Parameters
    ----------
    telescope: string
        Telescope identifier string. Currently supports jwst, roman, and hst.
    instrument: string
        Instrument identifier string.
    mode: string
        Mode identifier string.

    Returns
    -------
    calc: dict
        Engine output API compliant dict defining a calculation
    """
    calc = dict()
    ex_args = dict(**kwargs)

    ins_config = get_instrument_config(telescope=telescope, instrument=instrument)
    if mode is None or mode not in ins_config['modes']:
        mode = ins_config['defaults']['mode']

    # the instrument configuration now contains instrument and strategy values
    calc['configuration'] = ins_config['defaults'][mode]['instrument_args']

    calc['scene'] = build_default_scene(telescope)

    bg_defaults_file = "defaults/background_defaults.json"
    bg_defaults = read_json(pkg_resources.resource_filename("pandeia.engine", bg_defaults_file))

    calc['background'] = bg_defaults[telescope]['bg_location']
    calc['background_level'] = bg_defaults[telescope]['bg_level']

    calc_defaults_file = "defaults/calculationconfig_defaults.json"
    calc['calculation'] = read_json(pkg_resources.resource_filename("pandeia.engine", calc_defaults_file))

    st_defaults = ins_config['strategy_config'][mode]
    calc['strategy'] = dict()

    if 'method' in ex_args:
        calc['strategy']['method'] = ex_args['method']
    else:
        calc['strategy']['method'] = st_defaults['method']

    # Only use the constructed strategy with the default if this IS the default strategy.
    if ins_config['defaults'][mode]['strategy_args']['method'] == calc['strategy']['method']:
        calc['strategy'] = ins_config['defaults'][mode]['strategy_args']
    else:
        # otherwise, construct the entire strategy
        method_defaults_file = "%s.json" % calc['strategy']['method']

        method_defaults = read_json(os.path.join(default_refdata_directory, "strategy", method_defaults_file))
        required = method_defaults.pop("required")
        method_defaults.update(required)
        defaults = method_defaults.pop("defaults")
        method_defaults.update(defaults)
        method_defaults.pop("api", None)
        for k in method_defaults:
            if 'permitted' not in k:
                calc['strategy'][k] = method_defaults[k]

        # sometimes aperture_key is a dict because some strategies will use it, but others won't.
        # if it's not used, it will be set to None
        if 'aperture_key' in st_defaults:
            if isinstance(st_defaults['aperture_key'], dict):
                ap_key = st_defaults['aperture_key'][calc['strategy']['method']]
            else:
                ap_key = st_defaults['aperture_key']
            if ap_key is not None:
                key = calc['configuration']['instrument'][ap_key]
                calc['strategy']['aperture_size'] = st_defaults['aperture_sizes'][key]
        else:
            ap_key = None
            if 'aperture_size' in st_defaults:
                calc['strategy']['aperture_size'] = st_defaults['aperture_size']

        # in most cases sky_key will be the same as ap_key.  however, in the case of IFUNodApPhot, aperture
        # is used but NOT sky_annulus.  this is here to deal with that scenario, but may come up in future
        # scenarios as well.  if sky_annulus is not to be used, then sky_key will be set to None.

        #  NOTE this should get refactored to be more generalized.  use of sky_key for coronagraphy
        #  parameters is hackish... Also note this is now only used for IFU and MSA modes
        if "sky_key" in st_defaults:
            sky_key = st_defaults['sky_key'][calc['strategy']['method']]
        else:
            sky_key = ap_key

        # make sure sky_key is set and sky_annulus_sizes are actually defined
        params = {
            "sky_annulus_sizes": "sky_annulus",
            "unocculted_xys": "unocculted_xy",
            "contrast_azimuths": "contrast_azimuth",
            "contrast_separations": "contrast_separation"
        }
        for p in params:
            if sky_key is not None and p in st_defaults:
                key = calc['configuration']['instrument'][sky_key]
                calc['strategy'][params[p]] = st_defaults[p][key]

    calc = apply_default_only(calc)

    return calc


def calcs_to_wb(calc_list, name="Generated workbook", note="", desc="", wb_id=1, proposal_id="", proposal_state=""):
    """
    Take a list of calculations and make a workbook out of them.  This routine doesn't do any redundancy checking
    between scenes and sources so each calculation has its own scene with its own distinct sources.

    Parameters
    ----------
    calc_list: list of dicts (engine input API)
        List of dicts in engine input API compatible format

    Returns
    -------
    wb: dict
        Workbook data in pandeia workbook format
    """
    wb = {}
    scenes = {}
    sources = {}
    source_i = 1
    calcs = {}
    calc_i = 1
    for c in calc_list:
        # set up the scene for this calculation
        scenes[str(calc_i)] = {}
        scenes[str(calc_i)]["deleted"] = 0
        scenes[str(calc_i)]["characteristics"] = "{}"
        scenes[str(calc_i)]["name"] = "Scene #%d" % calc_i
        scenes[str(calc_i)]["desc"] = "Scene for calculation #%d" % calc_i
        scenes[str(calc_i)]["scene_elements"] = {}
        for src in c['scene']:
            # go through the scene and add the sources there to the workbook's list of sources
            scenes[str(calc_i)]["scene_elements"][str(source_i)] = src['position']
            scenes[str(calc_i)]["scene_elements"][str(source_i)]['name'] = "%d Calculation %s" % (source_i, calc_i)
            scenes[str(calc_i)]["scene_elements"][str(source_i)]['deleted'] = 0
            scenes[str(calc_i)]["scene_elements"][str(source_i)]['source_id'] = source_i
            scenes[str(calc_i)]["scene_elements"][str(source_i)]['source_arg_blob'] = []
            sources[str(source_i)] = {}
            sources[str(source_i)]['deleted'] = 0
            sources[str(source_i)]['name'] = "%f %s %s source" % (
                src['spectrum']['normalization']['norm_flux'],
                src['spectrum']['normalization']['norm_fluxunit'],
                src['shape']['geometry']
            )
            sources[str(source_i)]['characteristics'] = {}
            sources[str(source_i)]['characteristics']['shape'] = src['shape']
            sources[str(source_i)]['characteristics']['spectrum'] = src['spectrum']
            source_i += 1
        # set up the calculations
        calcs[str(calc_i)] = {}
        calcs[str(calc_i)]["strategy_args"] = c['strategy']
        calcs[str(calc_i)]["name"] = "Calculation #%d" % calc_i
        calcs[str(calc_i)]["deleted"] = 0
        calcs[str(calc_i)]["apt"] = 0
        calcs[str(calc_i)]["strategy"] = c['strategy']['method']
        calcs[str(calc_i)]["client_data"] = {
            "cause": "generated_wb"
        }
        calcs[str(calc_i)]["background"] = {"bg_type": c['background'], "ra": 0.0, "dec": 0.0}
        calcs[str(calc_i)]["scene_id"] = calc_i
        calcs[str(calc_i)]["camera_config"] = c['configuration']
        calcs[str(calc_i)]["instrument"] = calcs[str(calc_i)]["camera_config"]["instrument"].pop("instrument")
        calcs[str(calc_i)]["insmode"] = calcs[str(calc_i)]["camera_config"]["instrument"].pop("mode")
        calc_i += 1
    # now populate the workbook
    wb["test_mode"] = "0"
    wb["created"] = 0
    wb["deleted"] = 0
    wb["id"] = wb_id
    wb["name"] = name
    wb["desc"] = desc
    wb["note"] = note
    wb["proposal_id"] = proposal_id
    wb["proposal_state"] = proposal_state
    wb["calculations"] = calcs
    wb["scenes"] = scenes
    wb["sources"] = sources
    return wb



"""This module defines functions for converting BIT queries into pyetc
engine inputs.
"""
import os, pprint, re, sys

from   pandeia.engine.helpers.bit.pyetc_util import log, read_dict
import pandeia.engine.helpers.bit.pyetc_form_defaults as pyetc_defaults

# ========================================================================

def query_to_web_inputs(query_str):
    """Translate a single query string into a dictionary of partial
    JETC web inputs and a list of requested JETC result variables.

    query_str --> parital_web_inputs, requested_jetc_results
    """
    query_dict = query_to_dict(query_str)
    bit_inputs, requests = partition_query_dict(query_dict)
    web_inputs = expand_bit_inputs(bit_inputs)
    return web_inputs, requests

# ========================================================================

def query_to_dict(query):
    """Converts a web query GET parameter string into a Python dict mapping
    strings to strings.

    >>> query_to_dict("a=1&b=2&c=3")
    {'a': '1', 'c': '3', 'b': '2'}
    """
    keyvals = query.strip().split("&")
    kvpairs = []
    for assignment in keyvals:
        kvt = tuple([x.strip() for x in assignment.split("=")])
        if len(kvt) == 1:
            kvt = kvt + ("",)
        kvpairs.append(tuple(kvt))
    return dict(sorted(kvpairs))

def partition_query_dict(query_dict):
    """query_dict -->  (bit_inputs,  requests)"""
    bit_inputs = dict(query_dict)
    requests = []
    request_count = -1
    bit_inputs_keys = sorted(bit_inputs.keys())
    for key in bit_inputs_keys:
        if key.startswith("request"):
            if key == "requestCount":
                request_count = int(bit_inputs.pop(key))
            else:
                requests.append(bit_inputs.pop(key))
    assert len(requests) == request_count, "wrong number of requests"
    return bit_inputs, requests

# ========================================================================

def expand_bit_inputs(bit_inputs, config=None):
    """Expand `bit_inputs` into pyetc web inputs using `config` spec.   If
    `config` is None,  get the spec based on the "config" query variable.

    Return pyetc web inputs.
    """
    result = dict()
    if config is None:
        config = get_bit_config(bit_inputs["config"])
    for key, val in config.items():
        if isinstance(val, (str, int, float, bool)):
            result[key] = str(val)
        elif isinstance(val, dict) and key in bit_inputs:
            case = bit_inputs[key]
            for cases, what_to_do in val.items():
                if case in cases:
                    result.update(expand_bit_inputs(bit_inputs, what_to_do))
                    break
            else:
                raise ValueError("No handler found for " + key + " = " +
                                 repr(bit_inputs[key]))
        else:
            raise RuntimeError("Specification error at " + repr((key,val)))
    expanded = subst_vars(bit_inputs, result)
    return expanded

HERE = os.path.dirname(__file__) or "./"

def load_bit_config(config):
    """Load a single BIT configuration file combined with global settings.
    """
    try:
        path = HERE + "/config/global.dict"
        conf = read_dict(path)
        path = HERE + "/config/" + config.replace("/","_").lower() + ".dict"
        conf.update(read_dict(path))
    except IOError:
        raise ValueError("Unknown query config = " + repr(config))
    return conf

CONFIG_CACHE = {}
def get_bit_config(config):
    """Get a query expansion spec for a particular `config` setting,  looking
    first in the CONFIG_CACHE and loading the `config` only if not found.

    >>> _ = get_bit_config("acs/wfc")
    >>> _ = get_bit_config("acs/sbc")
    >>> _ = get_bit_config("cos/nuv")
    >>> _ = get_bit_config("cos/fuv")
    >>> _ = get_bit_config("wfc3/ir")
    >>> _ = get_bit_config("wfc3/uvis")
    >>> _ = get_bit_config("stis/ccd")
    >>> _ = get_bit_config("stis/nuv-mama")
    >>> _ = get_bit_config("stis/fuv-mama")
    """
    if config not in CONFIG_CACHE:
        CONFIG_CACHE[config] = load_bit_config(config)
    return CONFIG_CACHE[config]

LC_PERC_VAR = re.compile("%%([A-Za-z0-9_]+)%%")
PERC_VAR = re.compile("%([A-Za-z0-9_]+)%")

def subst_var(bit_inputs, string):
    """Replace %var% in `string` with bit_inputs[var].
    Replace %%var%% in `string` with bit_inputs[var].lower().
    Otherwise return `string`.
    """
    match = LC_PERC_VAR.search(string)
    if match:
        return LC_PERC_VAR.sub(bit_inputs[match.group(1)].lower(),
                           string)
    match = PERC_VAR.search(string)
    if match:
        return PERC_VAR.sub(bit_inputs[match.group(1)],
                            string)
    return string

def subst_vars(bit_inputs, config):
    """Substitute values from `bit_inputs` into all strings in BIT `config`,
    whether they be keys, values, simple strings, elements of lists, etc.
    """
    if isinstance(config, str):
        result = subst_var(bit_inputs, config)
    elif isinstance(config, (list, tuple)):
        result = []
        for elem in config:
            result.append(subst_vars(bit_inputs, elem))
        result = tuple(result)
    elif isinstance(config, dict):
        result = {}
        for key, value in config.items():
            result[subst_vars(bit_inputs, key)] = subst_vars(bit_inputs, value)
    elif isinstance(config, (int, float, bool)):
        result = str(config)
    else:
        raise TypeError("Don't know how to handle " + repr(config))
    return result

# ========================================================================

def convert_web_inputs(web_inputs):
    """Converts partial pyetc web inputs into fully specified pyetc
    web inputs.
    """
    instrument, sciencemode = get_migration_instr_sciencemode(web_inputs)

    web_inputs["instrument"] = web_inputs["instrument"].lower()
    web_inputs["science_mode"] = web_inputs["science_mode"].lower()

    # Get defaults for this instrument and sciencemode from the scraped
    # values kept by the spider tools.
    converted = pyetc_defaults.get_defaults(instrument, sciencemode)
    converted.update(web_inputs)

    # if instrument == "cos":
    #     hack_cos_inputs(converted)

    return converted

def hack_cos_inputs(jraw_inputs):
    """Do in-place hacks to fix wavelength issues for COS centralWavelength
    and obswave.
    """
    # This j2p hack may belong in the migration tools which are currently
    # supplying the obsolete JETC default of 1230.
    g140l = jraw_inputs.get("G140L_CentralWavelength", None)
    if g140l == "1230":
        jraw_inputs["G140L_CentralWavelength"] = "1280"

    # Ordinarily,  obswave is assigned the same value as centralWavelength from the query.
    # Sometimes it doesn't work so we hack just the required points until the full set of
    # ranges is implemented somewhere in Python.
    # RangeError("Wavelength 1105.0 not found in available cos segment ranges {(1120.0, 2250.0): 'a'}",)
    # RangeError("Wavelength 1280.0 not found in available cos segment ranges {(1281.0, 2385.0): 'a', (700.0, 1195.5): 'b'}",)
    # RangeError("Wavelength 1300.0 not found in available cos segment ranges {(1146.0, 1287.0): 'b', (1301.0, 1442.0): 'a'}",)
    # RangeError("Wavelength 1600.0 not found in available cos segment ranges {(1410.0, 1581.0): 'b', (1601.0, 1772.0): 'a'}",)
    hacked_obswaves = {
        "1105" : "1120",
        "1280" : "1281",
        "1300" : "1301",
        "1600" : "1601",
    }
    if "obswave" in jraw_inputs:
        obswave = jraw_inputs.get("obswave")
        jraw_inputs["obswave"] = hacked_obswaves.get(obswave, obswave)

def get_migration_instr_sciencemode(web_inputs):
    """Determine the instrument and science_mode with the spelling used by
    the spider tools.
    """
    instrument = web_inputs["instrument"].lower()
    sciencemode = {
        "Imaging" : "imag",
        "Spectroscopic" : "spec",
        "Rampfilter" : "ramp"
    }.get(web_inputs["science_mode"])
    return instrument, sciencemode

# ========================================================================

# Function to help analyze queries to determine parameter space.

def get_query_values(filename):
    """Utility to return the possible values for each variable defined in
    some query in `filename`.   For development analysis.

    queries(filename) -->  { var : [values...], ... }
    """
    values = {}
    for query in queries(filename):
        query_dict = query_to_dict(query)
        for key, value in query_dict.items():
            if key not in values:
                values[key] = [value]
            elif value not in values[key]:
                values[key].append(value)
    return values

# ========================================================================

# The main translation function called by BIT computations.

def process_query(query_str, info=True):
    """Translate and dump a single query.

    Returns {web_inputs}, [requested_outputs]
    """
    if info:
        log.info("BIT Query:", repr(query_str))
    web_inputs, requests = query_to_web_inputs(query_str)
    complete_web_inputs = convert_web_inputs(web_inputs)
    if log.VERBOSE_FLAG:
        log.verbose("\nComplete pyetc web inputs:\n",
                    pprint.pformat(complete_web_inputs))
    return complete_web_inputs, requests


# ========================================================================

# A regression test function to determine that engine inputs haven't changed.

def process_file(fname):
    """Translate an entire file containing one query per line.
    Outputs to a like-named .qweb file:  { query : { web_inputs } }
    """
    ofname = os.path.splitext(fname)[0]+".qweb"
    ofile = open(ofname, "w+")
    ofile.write("{\n")
    for query_str in queries(fname):
        try:
            web_inputs, requests = process_query(query_str, info=True)
        except Exception:
            log.error("Exception in " + repr(query_str))
        else:
            web_inputs["bit_requests"] = requests
            ofile.write(repr(query_str) + " : " + repr(web_inputs) + ",\n\n")
    ofile.write("}\n")

def queries(filename):
    """Generator returning the query strings defined in `filename`,
    removing comments.
    """
    for line in open(filename):
        if not line.strip().startswith("//"):
            yield line.strip()



"""
    This module provides the mapping between instrument names and telescope names.

    An instrument must know which telescope it is associated with. This
    is so because instrument modules are set into subpackages for each
    telescope. Thus the generic code must know from which subpackage to
    import the instrument-specific code.

"""

TELESCOPES = {'acs':      'hst',
              'stis':     'hst',
              'cos':      'hst',
              'wfc3uvis': 'hst',
              'wfc3ir':   'hst',
              'nircam':   'jwst',
              'nirspec':  'jwst',
              'miri':     'jwst',
              'niriss':   'jwst',
              }





import ast
import io
import os
import numpy as np
from numpy import nan, inf, NaN, Inf
from collections import defaultdict, OrderedDict

def read_dict(fname):
    '''read a python dictionary from a file that was written
    with write_dict.
    '''
    f=open(fname,'r', encoding="utf-8")
    datastr = f.read()
    f.close()
    # convert DOS file to Unix - otherwise the eval will fail
    datastr = datastr.replace('\r','')
    try :
        datadict = safe_eval(datastr)
    except Exception as e:
        print('EXCEPTION:',e)
        print('cannot eval data in file ',fname)
        raise
    return datadict

# Taken from Python-2.7 and enhanced to support np.nan and np.inf
def safe_eval(node_or_string):
    """
    Safely evaluate an expression node or a string containing a Python
    expression.  The string or node provided may only consist of the following
    Python literal structures: strings, numbers, tuples, lists, dicts, booleans,
    and None.
    """
    _safe_names = {'None': None, 'True': True, 'False': False,
                   'nan':nan, 'NaN':np.NaN,
                   'inf':inf, 'Inf':np.Inf}
    if isinstance(node_or_string, str):
        node_or_string = ast.parse(node_or_string, mode='eval')
    if isinstance(node_or_string, ast.Expression):
        node_or_string = node_or_string.body
    def _convert(node):
        # Old code used to return instances of ast.Dict. 
        # New code sometimes retuns dicts disguised as lists 
        # of tuples (actuaally, ast.List and ast.Tuple), as 
        # the first element of the 'args' attribute of an 
        # ast.Call instace. 
        if isinstance(node, ast.Call):
            result = OrderedDict()
            if len(node.args) > 0:
                for element in node.args[0].elts:
                    name  = element.elts[0].s
                    v = element.elts[1]
                    if isinstance(v, ast.Str): 
                        value = v.s
                    elif isinstance(v, ast.Num):
                        value = v.n
                    else:
                        value = None
                    result[name] = value
            return result
        elif isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Tuple):
            return tuple(map(_convert, node.elts))
        elif isinstance(node, ast.List):
            return list(map(_convert, node.elts))
        elif isinstance(node, ast.Dict):
            return OrderedDict((_convert(k), _convert(v)) for k, v
                         in zip(node.keys, node.values))
        elif isinstance(node, ast.Name):
            if node.id in _safe_names:
                return _safe_names[node.id]
        elif isinstance(node, ast.BinOp) and \
             isinstance(node.op, (ast.Add, ast.Sub)) and \
             isinstance(node.right, ast.Num) and \
             isinstance(node.right.n, complex) and \
             isinstance(node.left, ast.Num) and \
             isinstance(node.left.n, int) and \
             isinstance(node.left.n, float):
            left = node.left.n
            right = node.right.n
            if isinstance(node.op, ast.Add):
                return left + right
            else:
                return left - right

        # when the expression contains a negative number,
        # under python 3 the negative number will be cast 
        # as an instance of UnaryOp with first operand of
        # type USub. 
        elif isinstance(node, ast.NameConstant):
            if sys.version_info[0] >= 3:
                return node.value
        elif isinstance(node, ast.UnaryOp) and \
             isinstance(node.op, ast.USub) and \
             isinstance(node.operand, ast.Num): 
            return -(node.operand.n) 

        raise ValueError('malformed string')

    return _convert(node_or_string)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import numpy as np
import numpy.ma as ma
from .custom_exceptions import EngineInputError, DataError

class ExposureSpec:
    """
    Parent class for all exposure times, MultiAccum and SingleAccum
    """
class ExposureSpec_MultiAccum(ExposureSpec):
    """
    Parent class for MultiAccum
    """

    def __init__(self, config={}, webapp=False, **kwargs):
        """
        Create a generic Exposure Specification.

        Inputs
        ------
        config: dict
            dictionary of detector configuration setups

        webapp: bool
            Switch to toggle strict API checking
            
        **kwargs: keyword/value pairs
            Additional configuration data
        """
        self.webapp = webapp

        # Required parameters
        self.readout_pattern = config["input_detector"]["readout_pattern"]
        self.subarray = config["input_detector"]["subarray"]
        self.ngroup = config["input_detector"]["ngroup"]
        self.nint = config["input_detector"]["nint"]
        self.nexp = config["input_detector"]["nexp"]
        self.tframe = config["subarray"]["default"][self.subarray]["tframe"]
        self.tfffr = config["subarray"]["default"][self.subarray]["tfffr"]
        if self.readout_pattern in config["subarray"] and self.subarray in config["subarray"][self.readout_pattern]:
            self.tframe = config["subarray"][self.readout_pattern][self.subarray]["tframe"]
            self.tfffr = config["subarray"][self.readout_pattern][self.subarray]["tfffr"]

        # these errors specifically name the properties by their engine 
        # internal name; the client will catch errors for the web interface 
        # and thus the engine will never trigger this for web users. 
        # (JETC-1957)
        if not isinstance(self.ngroup, int):
            raise EngineInputError("ngroups must be an integer, got {}".format(type(self.ngroup)))
        if not isinstance(self.nint, int):
            raise EngineInputError("nint must be an integer, got {}".format(type(self.nint)))
        if not isinstance(self.nexp, int):
            raise EngineInputError("nexp must be an integer, got {}".format(type(self.nexp)))

        # Optional parameters
        # These are defined by the Instrument's reference data as Instrument properties.
        self.nframe = config["readout_pattern"][self.readout_pattern]["nframe"]
        self.ndrop2 = config["readout_pattern"][self.readout_pattern]["ndrop2"]

        self.nprerej = config["detector_config"]["nprerej"]
        self.npostrej = config["detector_config"]["npostrej"]
        # Going from general to specific: start with the default reset values, then check the defaults, then check
        # the actual readout pattern for this subarray. Use the most appropriate reset values for this setup, otherwise
        # the defaults
        self.nreset1 = 1
        self.nreset2 = 1
        if "nreset1" in config["subarray"]["default"][self.subarray]:
            self.nreset1 = config["subarray"]["default"][self.subarray]["nreset1"]
            self.nreset2 = config["subarray"]["default"][self.subarray]["nreset2"]
        if self.readout_pattern in config["subarray"] and self.subarray in config["subarray"][self.readout_pattern]:
            if "nreset1" in config["subarray"][self.readout_pattern][self.subarray]:
                self.nreset1 = config["subarray"][self.readout_pattern][self.subarray]["nreset1"]
                self.nreset2 = config["subarray"][self.readout_pattern][self.subarray]["nreset2"]
        if "nreset1" in config["readout_pattern"][self.readout_pattern]:
            self.nreset1 = config["readout_pattern"][self.readout_pattern]["nreset1"]
            self.nreset2 = config["readout_pattern"][self.readout_pattern]["nreset2"]


        # These are never specified in our data, currently; they were always the default
        # value from the ExposureSpec __init__ signature.
        if "ndrop1" in config["readout_pattern"][self.readout_pattern]:
            self.ndrop1 = config["readout_pattern"][self.readout_pattern]["ndrop1"]
        else:
            self.ndrop1 = 0
        if "ndrop3" in config["readout_pattern"][self.readout_pattern]:
            self.ndrop3 = config["readout_pattern"][self.readout_pattern]["ndrop3"]
        else:
            self.ndrop3 = 0
        self.frame0 = False
        if "frame0" in config["subarray"]["default"][self.subarray]:
            self.frame0 = config["subarray"]["default"][self.subarray]["frame0"]
        if self.readout_pattern in config["subarray"] and self.subarray in config["subarray"][self.readout_pattern]:
            if "frame0" in config["subarray"][self.readout_pattern][self.subarray]:
                self.nreset1 = config["subarray"][self.readout_pattern][self.subarray]["frame0"]

        # If these are trivial, we don't have to define them.
        if "nsample" in config["readout_pattern"][self.readout_pattern]:
            self.nsample = config["readout_pattern"][self.readout_pattern]["nsample"]
            self.nsample_skip = config["readout_pattern"][self.readout_pattern]["nsample_skip"]
        else:
            self.nsample = 1
            self.nsample_skip = 0

        # Target acqs only use a subset of the groups
        if "ngroup_extract" in config["input_detector"]:
            self.ngroup_extract = config["input_detector"]["ngroup_extract"]

        self.get_times()

        # Derived quantities
        self.nramps = self.nint * self.nexp

    def get_times(self):
        """
        The time formulae are defined in Holler et al. 2021, JWST-STScI-006013-A. Note that we have to subtract the
        groups that are rejected (by the pipeline) from the measurement time (nprerej+npostrej). The saturation time
        conservatively considers the ramp saturated even if saturation occurs in a rejected frame.

        Also note that these equations are generic, suitable for both H2RG and SiAs detectors.

        The equations in this method are duplicated in the front-end
        (workbook.js, update_detector_time_labels function). Please ensure
        that changes to these equations are reflected there.
        """

        self.tgroup = self.tframe * (self.nframe + self.ndrop2)

        # MIRI measurement time for ngroups < 5 is now handled with dropframes
        # This reduces to Equation 3 for H2RG detectors and Equation 5 for SiAs detectors.
        self.measurement_time = self.nint * self.tframe * (self.ngroup - 1 - self.nprerej - self.npostrej) * (self.nframe + self.ndrop2)
        # Equation 4
        if self.frame0:
            self.measurement_time += 0.5 * self.nint * self.tframe * (self.nframe - 1)

        # Equation 1, which naturally simplifies to Equation 2 for SiAs detectors.
        self.exposure_time = (self.tfffr * self.nint) + self.tframe * (self.nreset1 + (self.nint - 1) * self.nreset2 +
                                self.nint * (self.ndrop1 + (self.ngroup - 1) * (self.nframe + self.ndrop2) +
                                self.nframe + self.ndrop3))

        # Equation 6, which reduces to Equation 7 for SiAs detectors.
        self.saturation_time = self.tframe * (self.ndrop1 + (self.ngroup - 1) * (self.nframe + self.ndrop2) + self.nframe)

        self.duty_cycle = self.measurement_time / self.exposure_time
        self.total_exposure_time = self.nexp * self.exposure_time
        self.exposures = self.nexp

        self.total_integrations = self.nexp * self.nint

class ExposureSpec_H2RG(ExposureSpec_MultiAccum):

    pass


class ExposureSpec_SiAs(ExposureSpec_MultiAccum):

    def get_times(self):
        super().get_times()

        # This is where we adjust values so we can still use the same
        # MULTIACCUM formula as for the NIR detectors. We need the effective
        # "average time per sample" for MIRI.
        self.tsample = self.tframe / (self.nsample + self.nsample_skip)
        # 'nsample_total' for MIRI is the total number of non-skipped samples X number of averaged frames.
        # Note that in practice, it currently never happens that both the number of samples and number of
        # averaged frames are >1 (i.e., no SLOWGRPAVG exists). However, this will deal with that situation,
        # should it occur.
        self.nsample_total = self.nframe * self.nsample


class ExposureSpec_UnevenMultiAccum(ExposureSpec):

    def __init__(self, config={}, webapp=False, **kwargs):
        """
        Create a generic Exposure Specification.

        Inputs
        ------
        config: dict
            dictionary of detector configuration setups

        webapp: bool
            Switch to toggle strict API checking
            
        **kwargs: keyword/value pairs
            Additional configuration data
        """
        self.webapp = webapp

        # Required parameters
        self.ma_table_name = config["input_detector"]["ma_table_name"].lower()
        self.subarray = config["input_detector"]["subarray"]
        self.nresultants = config["input_detector"]["nresultants"]
        self.nexp = config["input_detector"]["nexp"]
        
        try:
            self.ma_table = config["readout_pattern"][self.ma_table_name]
        except KeyError:
            raise EngineInputError(f"Invalid MA Table name: {self.ma_table_name}")

        self.tframe = self.ma_table["frame_time"]
        self.treset = self.ma_table["reset_read_time"]
        self.nreset = self.ma_table["reset_reads"]
        self.nreference = 1 if self.ma_table["subtract_reference_read"] else 0

        self.max_total_samples = self.ma_table["max_samples"]
        self.max_resultants = self.ma_table["max_resultants"]
        self.min_resultants = self.ma_table["min_resultants"]
        # Allow a simple "maximum number of resultants" option
        if self.nresultants == -1:
            self.nresultants = self.max_resultants
        # Check, otherwise, for invalid resultant values.
        if self.nresultants > self.max_resultants:
            raise EngineInputError(f"MA Table {self.ma_table_name} supports a maximum of {self.max_resultants} resultants.")
        if self.nresultants < self.min_resultants:
            raise EngineInputError(f"MA Table {self.ma_table_name} supports a minimum of {self.min_resultants} resultants.")

        # we've read the full table, now consider only out to the requested stopping
        # point.
        try:
            self.readout_pattern = self._enumerate_pattern()[:self.nresultants]
        except TypeError:
            raise EngineInputError(f"Number of resultants ({self.nresultants}) must be an integer.")

        # Get number of reads in each resultant
        self.nreads = np.array([len(i) for i in self.readout_pattern])
        # Get total number of frames (read and unread) in each resultant
        first = np.array([i[0] for i in self.readout_pattern ])
        self.ntotal = np.append(np.diff(first), len(self.readout_pattern[-1]))
        self.lastframe = [i[-1] for i in self.readout_pattern]
        self.max_samples = self.readout_pattern[-1][-1]


        self.get_times()

    def _enumerate_pattern(self):
        """
        Given an MA Table pattern description, write out the list of resultants and frames in each resultant

        Returns
        -------
        readout_pattern: list
            A list of resultants, where each resultant is a list of consitituent frames.
        """
        resultants = self.ma_table["resultants"]
        frame_counter = 1
        readout_pattern = []
        for idx, resultant in enumerate(resultants):
            frames = frame_counter + resultant["pre_resultant_skips"] + np.arange(resultant["read_frames"])
            readout_pattern.append(list(frames))
            frame_counter += resultant["pre_resultant_skips"] + resultant["read_frames"]
        
        return readout_pattern

    def get_times(self):

        # The frame-based time between the middle of the first and middle of the last readout
        if self.nreference > 0:
            self.measurement_time = (self.nreference/2 + np.sum(self.ntotal[0:-1]) + self.ntotal[-1]/2) * self.tframe
        else:
            self.measurement_time = (self.ntotal[0] + np.sum(self.ntotal[1:-1]) + self.ntotal[-1]/2) * self.tframe

        # The total time spent on one exposure, including resets
        # Total number of frames + optional reference read + reset read
        self.exposure_time = 1 * self.treset + self.nreference * self.tframe + self.readout_pattern[-1][-1] * self.tframe

        # The total time spent exposing this observation (exposure time, times number of exposures)
        self.total_exposure_time = self.exposure_time * self.nexp

        # The time that's important for computing saturation; the time spent collecting photons.
        self.saturation_time = self.nreference * self.tframe + self.readout_pattern[-1][-1] * self.tframe

        self.duty_cycle = self.saturation_time/self.exposure_time
        self.nramps = self.nexp
        self.exposures = self.nexp
        self.total_integrations = self.nexp

class ExposureSpec_H4RG(ExposureSpec_UnevenMultiAccum):

    pass

class ExposureSpec_SingleAccum(ExposureSpec):
    """
    Parent class for SingleAccum
    """

    def __init__(self, config={}, webapp=False, **kwargs):
        """
        Create a single accum Exposure Specification.

        Inputs
        ------
        config: dict
            dictionary of detector configuration setups

        webapp: bool
            Switch to toggle strict API checking

        **kwargs: keyword/value pairs
            Additional configuration data

        """
        self.webapp = webapp

        self.time = config["input_detector"]["time"]

        # Required parameters
        #self.readout_pattern = config["input_detector"]["readout_pattern"]
        #self.subarray = config["input_detector"]["subarray"]
        if "nexp" in config["input_detector"]:
            raise DataError("SingleAccum calculations cannot use nexp")
        self.nsplit = config["input_detector"]["nsplit"]

        self.get_times()

        # Derived quantities needed for the generic noise equation
        self.nramps = 1

    def get_times(self):
        # See also SingleAccum.set_time()

        # measurement time is the time from the first measurement to the last read of the
        # exposure. HST has no skipped or dropped frames; this is simply the time for one
        # frame.
        self.measurement_time = self.time / self.nsplit

        # exposure time is the total time of one exposure (including any skipped reads,
        # reset reads, and the like - HST doesn’t have them.)
        self.exposure_time = self.time / self.nsplit

        # saturation time is the time the detector spends collecting photons between
        # resets, including skipped and dropped reads. It is the time that saturation
        # calculations depend on.
        self.saturation_time = self.time / self.nsplit

        # total exposure time is the time for the entire observation, including all
        # exposures and all the resets and skipped frames.
        self.total_exposure_time = self.time

        self.duty_cycle = self.saturation_time/self.exposure_time
        self.total_integrations = self.nsplit
        self.exposures = self.nsplit


class ExposureSpec_CCD(ExposureSpec_SingleAccum):

    pass

class ExposureSpec_H1R(ExposureSpec_CCD):

    pass

class ExposureSpec_MAMA(ExposureSpec_SingleAccum):

    pass

class ExposureSpec_XDL(ExposureSpec_MAMA):

    pass



#!/usr/bin/env python

# setuptools is required
from setuptools import setup

import subprocess

with open('README.md') as fp:
    description = fp.read()

# Versioning
try:
    DEVBUILD = subprocess.check_output(["git", "describe", "--tags"])
    with open('pandeia/engine/helpers/DEVBUILD', 'wb') as out:
        out.write(DEVBUILD)
except (subprocess.CalledProcessError, FileNotFoundError) as err:
    print(err)

setup(
    # The package
    name="pandeia.engine",
    version="3.1",
    packages=["pandeia",
              "pandeia.engine",
              "pandeia.engine.defaults",
              "pandeia.engine.helpers",
              "pandeia.engine.helpers.bit",
              "pandeia.engine.helpers.bit.config",
              "pandeia.engine.helpers.bit.etc_web_data",
              "pandeia.engine.helpers.bit.pyetc_form_defaults",
              "pandeia.engine.helpers.bit.instruments",
              "pandeia.engine.helpers.bit.instruments.hst",
              "pandeia.engine.helpers.bit.instruments.hst.acs",
              "pandeia.engine.helpers.bit.instruments.hst.acs.web",
              "pandeia.engine.helpers.bit.instruments.hst.cos",
              "pandeia.engine.helpers.bit.instruments.hst.cos.web",
              "pandeia.engine.helpers.bit.instruments.hst.stis",
              "pandeia.engine.helpers.bit.instruments.hst.stis.web",
              "pandeia.engine.helpers.bit.instruments.hst.wfc3ir",
              "pandeia.engine.helpers.bit.instruments.hst.wfc3ir.web",
              "pandeia.engine.helpers.bit.instruments.hst.wfc3uvis",
              "pandeia.engine.helpers.bit.instruments.hst.wfc3uvis.web",
              "pandeia.engine.helpers.peng",
              "pandeia.engine.helpers.schema",
              "pandeia.engine.helpers.background",
              "pandeia.engine.helpers.background.hst",
              "pandeia.engine.helpers.background.jwst"],

    # For PyPI
    description='Pandeia 3D Exposure Time Calculator compute engine',
    long_description=description,
    author='Adric Riedel, Isaac Spitzer, Dharini Chittiraibalan, Oi In Tam, Chris Sontag, Ivo Busko, Craig Jones, Tim Pickering, Klaus Pontoppidan',
    #author_email='https://stsci.service-now.com/jwst',
    url='https://jwst.etc.stsci.edu',
    classifiers=["Intended Audience :: Science/Research",
                 "License :: OSI Approved :: BSD License",
                 "Operating System :: OS Independent",
                 "Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3.8",
                 "Programming Language :: Python :: 3.9",
                 "Programming Language :: Python :: 3.10",
                 "Programming Language :: Python :: 3.11",
                 "Topic :: Scientific/Engineering :: Astronomy",
                 "Topic :: Software Development :: Libraries :: Python Modules"],
    # Other notes
    package_data={
        "pandeia.engine.defaults": ["*.json"],
        "pandeia.engine.helpers": ["DEVBUILD"],
        "pandeia.engine.helpers.bit.instruments.hst.stis":     ["*.dat"],
        "pandeia.engine.helpers.bit.instruments.hst.cos":      ["*.dat"],
        "pandeia.engine.helpers.bit.instruments.hst.acs":      ["*.dat"],
        "pandeia.engine.helpers.bit.instruments.hst.wfc3uvis": ["*.dat"],
        "pandeia.engine.helpers.bit.instruments.hst.wfc3ir":   ["*.dat"],
        },
    include_package_data=True,
    install_requires=[
        "numpy>=1.17",
        "scipy>1.6",
        "astropy>=4",
        "photutils",
        "synphot",
        "stsynphot",
        "setuptools"
    ],
    zip_safe=False
)
